/*    [-*- C++ -*-] */
/*
THE COMPUTER CODE CONTAINED HEREIN IS THE SOLE PROPERTY OF PARALLAX
SOFTWARE CORPORATION ("PARALLAX").  PARALLAX, IN DISTRIBUTING THE CODE TO
END-USERS, AND SUBJECT TO ALL OF THE TERMS AND CONDITIONS HEREIN, GRANTS A
ROYALTY-FREE, PERPETUAL LICENSE TO SUCH END-USERS FOR USE BY SUCH END-USERS
IN USING, DISPLAYING,  AND CREATING DERIVATIVE WORKS THEREOF, SO LONG AS
SUCH USE, DISPLAY OR CREATION IS FOR NON-COMMERCIAL, ROYALTY OR REVENUE
FREE PURPOSES.  IN NO EVENT SHALL THE END-USER USE THE COMPUTER CODE
CONTAINED HEREIN FOR REVENUE-BEARING PURPOSES.  THE END-USER UNDERSTANDS
AND AGREES TO THE TERMS HEREIN AND ACCEPTS THE SAME BY USE OF THIS FILE.  
COPYRIGHT 1993-1998 PARALLAX SOFTWARE CORPORATION.  ALL RIGHTS RESERVED.
*/
/*
 * $Source: Smoke:miner:source:texmap::RCS:scanline.c $
 * $Revision: 1.4 $
 * $Author: allender $
 * $Date: 1995/08/14 14:28:28 $
 * 
 * Routines to draw the texture mapped scanlines.
 * 
 * $Log: scanline.c $
 * Revision 1.4  1995/08/14  14:28:28  allender
 * changed transparency color to 0
 *
 * Revision 1.3  1995/07/05  16:22:47  allender
 * all asm verison when necessary
 *
 * Revision 1.2  1995/05/11  12:59:05  allender
 * mike subdivision, plus change transparency color
 *
 * Revision 1.1  1995/05/04  20:14:19  allender
 * Initial revision
 *
 * Revision 1.2  1995/02/20  18:23:39  john
 * Added new module for C versions of inner loops.
 * 
 * Revision 1.1  1995/02/20  17:42:27  john
 * Initial revision
 * 
 * 
 */



#include <math.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>

#include "fix.h"
#include "mono.h"
#include "gr.h"
#include "grdef.h"
#include "texmap.h"
#include "texmapl.h"
#include "scanline.h"



//#define USE_OLD_TMAP_SCANLINE

//WHS: divide table
//extern int divtab[]; //see ntmap.c (max z gezien: 315000 tegen muur)
#define DIV_TABLE_4
//#define DIV_TABLE_1


#ifdef DIV_TABLE_4
#  define DIVTAB_SIZE_0 (8192+0)
#  define DIVTAB_SIZE_1 (2048+0)
#  define DIVTAB_SIZE_2 (1024+0)
#  define DIVTAB_SIZE_3 (512+0)
#  define DIVTAB_SIZE_4 (768+0)
#  define DIVTAB_STEP_0 (2)
#  define DIVTAB_STEP_1 (8)
#  define DIVTAB_STEP_2 (32)
#  define DIVTAB_STEP_3 (128)
#  define DIVTAB_STEP_4 (512)
#  define DIVTAB_OFFSET_1 (DIVTAB_SIZE_0)
#  define DIVTAB_OFFSET_2 (DIVTAB_OFFSET_1+DIVTAB_SIZE_1)
#  define DIVTAB_OFFSET_3 (DIVTAB_OFFSET_2+DIVTAB_SIZE_2)
#  define DIVTAB_OFFSET_4 (DIVTAB_OFFSET_3+DIVTAB_SIZE_3)
#  define DIVTAB_SIZE (DIVTAB_OFFSET_4+DIVTAB_SIZE_4)
#endif

#ifdef DIV_TABLE_3
#  define DIVTAB_SIZE_0 (4096+0)
#  define DIVTAB_SIZE_1 (1536+0)
#  define DIVTAB_SIZE_2 (768+0)
#  define DIVTAB_STEP_0 (8)
#  define DIVTAB_STEP_1 (64)
#  define DIVTAB_STEP_2 (512)
#  define DIVTAB_OFFSET_1 (DIVTAB_SIZE_0)
#  define DIVTAB_OFFSET_2 (DIVTAB_OFFSET_1+DIVTAB_SIZE_1)
#  define DIVTAB_SIZE (DIVTAB_OFFSET_2+DIVTAB_SIZE_2)
#endif


#ifdef DIV_TABLE_1 /* less accurate */
#  define DIVTAB_SIZE_0 (2048+1)
#  define DIVTAB_SIZE_1 (768+1)
#  define DIVTAB_STEP_0 (64)
#  define DIVTAB_STEP_1 (512)
#  define DIVTAB_OFFSET_1 (DIVTAB_SIZE_0)
#  define DIVTAB_SIZE (DIVTAB_OFFSET_1+DIVTAB_SIZE_1)
#endif

int divtab[DIVTAB_SIZE];



/* this function is hardly called */
void c_tmap_scanline_flat()
{
	ubyte *dest;
	int x;

	dest = (ubyte *)(write_buffer + fx_xleft + (bytes_per_row * fx_y )  );

	for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
		*dest++ = tmap_flat_color;
	}

/* faster? not really (see profiling data) */
//	memset(dest, tmap_flat_color, fx_xright-fx_xleft+1);
}

void c_tmap_scanline_shaded()
{
	int fade;
	ubyte *dest;
	int x;

	dest = (ubyte *)(write_buffer + fx_xleft + (bytes_per_row * fx_y)  );

	fade = tmap_flat_shade_value<<8;
	for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
		*dest++ = gr_fade_table[ fade |(*dest)];
	}
}

void c_tmap_scanline_lin_nolight()
{
	ubyte *dest;
	uint c;
	int x;
	fix u,v,dudx, dvdx;

	u = fx_u;
	v = fx_v*64;
	dudx = fx_du_dx; 
	dvdx = fx_dv_dx*64; 

	dest = (ubyte *)(write_buffer + fx_xleft + (bytes_per_row * fx_y)  );

	if (!Transparency_on)	{
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			*dest++ = (uint)pixptr[ (f2i(v)&(64*63)) + (f2i(u)&63) ];
			u += dudx;
			v += dvdx;
		}
	} else {
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			c = (uint)pixptr[ (f2i(v)&(64*63)) + (f2i(u)&63) ];
			if ( c!=TRANSPARENCY_COLOR)
				*dest = c;
			dest++;
			u += dudx;
			v += dvdx;
		}
	}
}


void c_tmap_scanline_lin()
{
	ubyte *dest;
	uint c;
	int x, j;
	fix u,v,l,dudx, dvdx, dldx;

	u = fx_u;
	v = fx_v*64;
	dudx = fx_du_dx; 
	dvdx = fx_dv_dx*64; 

	l = fx_l>>8;
	dldx = fx_dl_dx>>8;
	dest = (ubyte *)(write_buffer + fx_xleft + (bytes_per_row * fx_y)  );

#if 0
	if (!Transparency_on)	{
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ (f2i(v)&(64*63)) + (f2i(u)&63) ] ];
			l += dldx;
			u += dudx;
			v += dvdx;
		}
	} else {
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			c = (uint)pixptr[ (f2i(v)&(64*63)) + (f2i(u)&63) ];
			if ( c!=TRANSPARENCY_COLOR)
				*dest = gr_fade_table[ (l&(0xff00)) + c ];
			dest++;
			l += dldx;
			u += dudx;
			v += dvdx;
		}
	}
#else
	if (!Transparency_on)	{
		ubyte*			pixPtrLocalCopy = pixptr;
		ubyte*			fadeTableLocalCopy = gr_fade_table;
		unsigned int	destlong;

		x = fx_xright-fx_xleft+1;

		if ((j = (unsigned int) dest & 3) != 0)
			{
			j = 4 - j;

			if (j > x)
				j = x;

			while (j > 0)
				{
				*dest++ = (unsigned int) fadeTableLocalCopy[ (l&(0xff00)) + (uint) pixPtrLocalCopy[ (f2i(v)&(64*63)) + (f2i(u)&63) ] ];
				l += dldx;
				u += dudx;
				v += dvdx;
				x--;
				j--;
				}
			}

		j &= ~3;
		while (j > 0)
			{
			destlong = (unsigned int) fadeTableLocalCopy[ (l&(0xff00)) + (uint) pixPtrLocalCopy[ (f2i(v)&(64*63)) + (f2i(u)&63) ] ] << 24;
			l += dldx;
			u += dudx;
			v += dvdx;
			destlong |= (unsigned int) fadeTableLocalCopy[ (l&(0xff00)) + (uint) pixPtrLocalCopy[ (f2i(v)&(64*63)) + (f2i(u)&63) ] ] << 16;
			l += dldx;
			u += dudx;
			v += dvdx;
			destlong |= (unsigned int) fadeTableLocalCopy[ (l&(0xff00)) + (uint) pixPtrLocalCopy[ (f2i(v)&(64*63)) + (f2i(u)&63) ] ] << 8;
			l += dldx;
			u += dudx;
			v += dvdx;
			destlong |= (unsigned int) fadeTableLocalCopy[ (l&(0xff00)) + (uint) pixPtrLocalCopy[ (f2i(v)&(64*63)) + (f2i(u)&63) ] ];
			l += dldx;
			u += dudx;
			v += dvdx;
			*((unsigned int *) dest) = destlong;
			dest += 4;
			x -= 4;
			j -= 4;
			}

		while (x-- > 0)
			{
			*dest++ = (unsigned int) fadeTableLocalCopy[ (l&(0xff00)) + (uint) pixPtrLocalCopy[ (f2i(v)&(64*63)) + (f2i(u)&63) ] ];
			l += dldx;
			u += dudx;
			v += dvdx;
			}

	} else {
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			c = (uint)pixptr[ (f2i(v)&(64*63)) + (f2i(u)&63) ];
			if ( c!=TRANSPARENCY_COLOR)
				*dest = gr_fade_table[ (l&(0xff00)) + c ];
			dest++;
			l += dldx;
			u += dudx;
			v += dvdx;
		}
	}
#endif
}


#define	MIN_SUBDIVIDABLE_LENGTH	4
#define	MAX_SUBDIVIDABLE_Z_RATIO (F1_0/4)

// -------------------------------------------------------------------------------------------------------------
//	Mike's subdividing scanline renderer.
void c_tmap_scanline_per_sub(int y, fix xleft, fix xright, fix uleft, fix uright, fix vleft, fix vright, fix zleft, fix zright, fix lleft, fix lright)
{
	int	dx = f2i(xright) - f2i(xleft);

	if (dx >= MIN_SUBDIVIDABLE_LENGTH) {
		int	subdivide = 0;

		if (zleft < zright) {
			if (zright - zleft < fixmul(MAX_SUBDIVIDABLE_Z_RATIO, zleft))
				subdivide = 1;
		} else if (zleft - zright < fixmul(MAX_SUBDIVIDABLE_Z_RATIO, zright))
			subdivide = 1;

		if (subdivide) {
			c_tmap_scanline_per_sub(y, xleft, (xright + xleft)/2, uleft, (uright + uleft)/2, vleft, (vright + vleft)/2, zleft, (zright + zleft)/2, lleft, (lright + lleft)/2);

			c_tmap_scanline_per_sub(y, (xright + xleft)/2, xright, (uright + uleft)/2, uright, (vright + vleft)/2, vright, (zright + zleft)/2, zright, (lright + lleft)/2, lright);
			return;
		}
	}


{
	ubyte *dest;
	uint c;
	int x;
	fix u,v,l,dudx, dvdx, dldx;

	u = fx_u;
	v = fx_v*64;
	dudx = fx_du_dx;
	dvdx = fx_dv_dx*64;

	l = fx_l>>8;
	dldx = fx_dl_dx>>8;
	dest = (ubyte *)(write_buffer + fx_xleft + (bytes_per_row * fx_y)  );

	if (!Transparency_on)	{
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ (f2i(v)&(64*63)) + (f2i(u)&63) ] ];
//			*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ (f2i(v)&(64*63)) | (f2i(u)&63) ] ];
			l += dldx;
			u += dudx;
			v += dvdx;
		}
	} else {
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			c = (uint)pixptr[ (f2i(v)&(64*63)) + (f2i(u)&63) ];
//			c = (uint)pixptr[ (f2i(v)&(64*63)) | (f2i(u)&63) ];

			if ( c!=TRANSPARENCY_COLOR)
				*dest = gr_fade_table[ (l&(0xff00)) + c ];
			dest++;
			l += dldx;
			u += dudx;
			v += dvdx;
		}
	}
}

}

void c_tmap_scanline_per_nolight()
{
	ubyte *dest;
	uint c;
	int x;
	fix u,v,z,dudx, dvdx, dzdx;

	u = fx_u;
//	v = fx_v*64;
	v = fx_v<<6;

	z = fx_z;
	dudx = fx_du_dx;
//	dvdx = fx_dv_dx*64;
	dvdx = fx_dv_dx<<6;

	dzdx = fx_dz_dx;

	dest = (ubyte *)(write_buffer + fx_xleft + (bytes_per_row * fx_y)  );

	if (!Transparency_on)	{
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
#ifdef USE_OLD_TMAP_SCANLINE
			*dest++ = (uint)pixptr[ ( (v/z)&(64*63) ) + ((u/z)&63) ];
#else
			{
				int f=(65536*256*4+(z>>1))/(z);
				*dest++ = (uint)pixptr[ ( ( (((v+32)>>6)*f)>>20)&(64*63) ) + ( ((u*f)>>26)&(63) ) ];
			}
#endif
			u += dudx;
			v += dvdx;
			z += dzdx;
		}
	} else {
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
#ifdef USE_OLD_TMAP_SCANLINE
			c = (uint)pixptr[ ( (v/z)&(64*63) ) + ((u/z)&63) ];
#else
			{
				int f=(65536*256*4+(z>>1))/(z);
				c = (uint)pixptr[ ( ( (((v+32)>>6)*f)>>20)&(64*63) ) | ( ((u*f)>>26)&(63) ) ];
			}
#endif
			if ( c!=TRANSPARENCY_COLOR)
				*dest = c;
			dest++;
			u += dudx;
			v += dvdx;
			z += dzdx;
		}
	}
}


 /* WHS: The first try in part 2 of tmap_scanline_per() gave me about 21 fps instead of 20 (then I tried part 1 which gets called a lot more), how much do the other routines above get called?), even on a ppro, divides are damn slow. */

/* See c_tmap_scanline_per_nolight() for the first version */

/* ntmap_scanline_per() speedup's:
 * V1: Too slow:
 *				long long f=(65536*16384)/(z);
 *				*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ (((v*f)>>30) & (64*63)) | (((u*f)>>30)&63) ] ];
 *
 *
 * V2: Tremendous speedup!! about 27-28 fps under X instead of 20 under the same circumstances (disadvantage: tile distortion close up)
 *				int f=(65536*256)/(z);
 *				*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( ( (((v+0)>>6)*f)>>18)&(64*63) ) | ( ((u*f)>>24)&(63) ) ] ];
 *
 *
 * V3: Use all 32 bits + rounding (needed?): (result: good; fixes tile distortion)
 *				int f=(65536*256*4+z-1)/(z);
 *
 *
 * V4:
 *				int f;
 * V4.1
 *				if (z>130000) f=(65536*256*4+(z>>1))/(z); else f=divtab[(z+32)>>6];
 * V4.2
 *				if (z>130000) f=divtab[((z+256)>>9) + DIVTAB_SIZE_1]; else f=divtab[(z+32)>>6];
 * V4.3
 *				if (z<131072) f=divtab[(z+32)>>6];
 *				else if (z<400000) f=divtab[((z+256)>>9) + DIVTAB_SIZE_1];
 * //				else f=(65536*256*4+(z>>1))/(z); //65536*256*4/512000
 *				else { dest++; goto skip_a; }
 * V4.X
 *				*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( ( (((v+63)>>6)*f)>>20)&(64*63) ) | ( ((u*f)>>26)&(63) ) ] ];
 */


#ifdef ASM_X86
#  define ASM_X86_TMAP_SCANLINE_PER
#  define ASM_TMAP_SCANLINE_PER
#endif
#ifdef ASM_PPC
#  define ASM_PPC_TMAP_SCANLINE_PER
#  define ASM_TMAP_SCANLINE_PER
#endif

#ifndef ASM_TMAP_SCANLINE_PER
void c_tmap_scanline_per()
{
	ubyte *dest;
	uint c;
	int x;
	fix u,v,z,l,dudx, dvdx, dzdx, dldx;

	u = fx_u;
	v = fx_v*64;
	z = fx_z;
	dudx = fx_du_dx;
	dvdx = fx_dv_dx*64;
	dzdx = fx_dz_dx;

	l = fx_l>>8;
	dldx = fx_dl_dx>>8;
	dest = (ubyte *)(write_buffer + fx_xleft + (bytes_per_row * fx_y)  );
	if (!Transparency_on)	{
			/* --- PART 1 --------------------------------------------- */

#ifdef EXP /* New experimental versions : No improvement */
#  define EXP_V2

#  ifdef EXP_V2 /* NB. USES DIVTABLE_1 */
#    ifndef DIV_TABLE_1
#      error need divtab 1
#    endif
		int f;
			x=fx_xright-fx_xleft+1   +1;
			if (x>0) {
			  loop_start:
				if (z>=320000) goto loop_too_close;
				if (z>=131072) goto loop_330k;
//x++;
			  loop_128k:
if (! --x) return;
//				if (z>=131072) goto loop_330k;
				f=divtab[z>>6];
//				if (z<131072) f=divtab[z>>6];
				*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( ( ((v>>6)*f)>>20)&(64*63) ) | ( ((u*f)>>26)&(63) ) ] ];
				l += dldx;
				u += dudx;
				v += dvdx;
				z += dzdx;

//				if (--x ==0) return;
				if (z<131072) goto loop_128k;
//goto loop_128k;
//if (!(z>>17)) goto loop_128k;
//if (--x ==0) return;
//if (--x) goto loop_128k;
//return;
			  loop_330k:
if (--x ==0) return;
				f=divtab[(z>>9) + DIVTAB_SIZE_1-256];
//if (z<320000) f=divtab[(z>>9) + DIVTAB_SIZE_1-256];
				*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( ( ((v>>6)*f)>>20)&(64*63) ) | ( ((u*f)>>26)&(63) ) ] ];

				l += dldx;
				u += dudx;
				v += dvdx;
				z += dzdx;

//				if (--x ==0) return;
				if (z<131072) goto loop_128k;
				else if (z<320000) goto loop_330k;

			  loop_too_close:
				if (--x ==0) return;
				dest++;

				l += dldx;
				u += dudx;
				v += dvdx;
				z += dzdx;

//				if (--x ==0) return;
				if (z>=320000) goto loop_too_close;
				else if (z>=131072) goto loop_330k;
				else goto loop_128k;
			}
#  endif /* EXP_V2 */
#else /* Older versions */

#ifdef USE_OLD_TMAP_SCANLINE /* keep to compare */
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( (v/z)&(64*63) ) + ((u/z)&63) ] ];
			l += dldx;
			u += dudx;
			v += dvdx;
			z += dzdx;
		}
#else
#  ifdef DIV_TABLE_4
//#    define INTERPOLATE
//#    define NEWv4_1 /* slow 36 ipv 45 */
//#    define NEWv4_2 /* ?? */
//#    define NEWv4_3
#    define NEWv4_4

#    ifdef INTERPOLATE /* new version that does less table lookups */
		{
		int f;
		int u_f, v_f;
		int u_f_next, v_f_next;
		int d_u_f, d_v_f;

		int i;

		v>>=6; dvdx>>=6;
dzdx<<=2; dudx<<=2; dvdx<<=2;
		if (z<16384)  f=divtab[z>>1];
		else if (z<32768)  f=divtab[(z>>3)+DIVTAB_OFFSET_1-2048];
		else if (z<65536)  f=divtab[(z>>5)+DIVTAB_OFFSET_2-1024];
		else if (z<131072) f=divtab[(z>>7)+DIVTAB_OFFSET_3- 512];
		else if (z<320000) f=divtab[(z>>9)+DIVTAB_OFFSET_4- 256];
		else { f=0; }

		v_f=(((uint) (v*f)) >>26) <<2;
		u_f=(((uint) (u*f)) >>26) <<2;

		for (x= (fx_xright-fx_xleft+1)>>2 ; x > 0; x-- ) {

//			z += dzdx<<2;
z+=dzdx;
			if (z<16384)  f=divtab[z>>1];
			else if (z<32768)  f=divtab[(z>>3)+DIVTAB_OFFSET_1-2048];
			else if (z<65536)  f=divtab[(z>>5)+DIVTAB_OFFSET_2-1024];
			else if (z<131072) f=divtab[(z>>7)+DIVTAB_OFFSET_3- 512];
			else if (z<320000) f=divtab[(z>>9)+DIVTAB_OFFSET_4- 256];
			else { f=0; }
//			u += dudx<<2;
//			v += dvdx<<2;
u +=dudx;
v +=dvdx;
			v_f_next = (((uint) (v*f)) >>26)<<2;
			d_v_f = v_f_next-v_f;
			u_f_next = (((uint) (u*f)) >>26)<<2;
			d_u_f = u_f_next-u_f;

			for (i=4; i; i--) {
				*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ((v_f<<4) &(63*64)) | (((u_f)>>2) &63) ] ];
				l += dldx;
				v_f = (v_f+d_v_f);
				u_f = (u_f+d_u_f);
			}
//			v_f=v_f_next;
//			u_f=u_f_next;

		}
#if 0
			u += dudx<<2;
			v += dvdx<<2;
			z += dzdx<<2;
			if (z<16384)  f=divtab[z>>1];
			else if (z<32768)  f=divtab[(z>>3)+DIVTAB_OFFSET_1-2048];
			else if (z<65536)  f=divtab[(z>>5)+DIVTAB_OFFSET_2-1024];
			else if (z<131072) f=divtab[(z>>7)+DIVTAB_OFFSET_3- 512];
			else if (z<320000) f=divtab[(z>>9)+DIVTAB_OFFSET_4- 256];
			else { f=0; }
			v_f_next = (((v>>6)*f)>>20)&(64*63);
			u_f_next = ((u*f)>>26)&(63);
			d_v_f = (v_f_next-v_f)>>2;
			d_u_f = (u_f_next-u_f)>>2;

			for (i=(fx_xright-fx_xleft+1)&3; i; i--) {
				*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ v_f | u_f ] ];
				l += dldx;
				u_f =(u_f+d_u_f)&63; v_f =(v_f+d_v_f)&(63*64);
			}
#endif
		}
#    elif defined(NEWv4_1)
		short s_l=(short) l;
		short s_dldx= (short) dldx;
//		short s_u=(short) u;
//		short s_dudx= (short) dudx;
//		short s_v=(short) v;
//		short s_dvdx= (short) dvdx;

		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			int f;
			if (z<16384)  f=divtab[z>>1];
			else if (z<32768)  f=divtab[(z>>3)+DIVTAB_OFFSET_1-2048];
			else if (z<65536)  f=divtab[(z>>5)+DIVTAB_OFFSET_2-1024];
			else if (z<131072) f=divtab[(z>>7)+DIVTAB_OFFSET_3- 512];
			else if (z<320000) f=divtab[(z>>9)+DIVTAB_OFFSET_4- 256];
			else { dest++; goto skip_a; }
			*dest++ = gr_fade_table[ (s_l&(0xff00)) + (uint)pixptr[ ( ( ((v>>6)*f)>>20)&(64*63) ) | ( ((u*f)>>26)&(63) ) ] ];
		  skip_a:
			s_l += s_dldx;
			u += dudx;
			v += dvdx;
			z += dzdx;
		}
#    elif defined(NEWv4_2)
uint v_u, d_v_u;
		v>>=6; dvdx>>=6;
		v_u=(v <<16) | (u&65535);
		d_v_u= (dvdx<<16) | (dudx &16383);

		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			int f;
			if (z<16384)  f=divtab[z>>1];
			else if (z<32768)  f=divtab[(z>>3)+DIVTAB_OFFSET_1-2048];
			else if (z<65536)  f=divtab[(z>>5)+DIVTAB_OFFSET_2-1024];
			else if (z<131072) f=divtab[(z>>7)+DIVTAB_OFFSET_3- 512];
			else if (z<320000) f=divtab[(z>>9)+DIVTAB_OFFSET_4- 256];
			else { dest++; goto skip_a; }
//			*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( ((v*f)>>20)&(64*63) ) | ( ((u*f)>>26)&(63) ) ] ];
			*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( (uint) ( (((v_u>>16)*f)&(63<<26) ) | ( ((uint)((v_u&65535)*f)>>6)) ) )>>20 ] ];

		  skip_a:
			l += dldx;
			v_u +=d_v_u;
			z += dzdx;
		}
#    elif defined(NEWv4_3)
		v>>=6; dvdx>>=6;

		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			int f;
			if (z<16384)  f=divtab[z>>1];
			else if (z<32768)  f=divtab[(z>>3)+DIVTAB_OFFSET_1-2048];
			else if (z<65536)  f=divtab[(z>>5)+DIVTAB_OFFSET_2-1024];
			else if (z<131072) f=divtab[(z>>7)+DIVTAB_OFFSET_3- 512];
			else if (z<320000) f=divtab[(z>>9)+DIVTAB_OFFSET_4- 256];
			else { dest++; goto skip_a; }
//			*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( ((v*f)>>20)&(64*63) ) | ( ((u*f)>>26)&(63) ) ] ];
			*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( (uint) ( ((v*f)&(63<<26) ) | ( ((uint)(u*f)>>6)) ) )>>20 ] ];

		  skip_a:
			l += dldx;
			u += dudx;
			v += dvdx;
			z += dzdx;
		}
#    elif defined(NEWv4_4)
uint aaa=63<<26;
		v>>=6; dvdx>>=6;

		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			int f;
			if (z<16384)  f=divtab[z>>1];
			else if (z<32768)  f=divtab[(z>>3)+DIVTAB_OFFSET_1-2048];
			else if (z<65536)  f=divtab[(z>>5)+DIVTAB_OFFSET_2-1024];
			else if (z<131072) f=divtab[(z>>7)+DIVTAB_OFFSET_3- 512];
			else if (z<340000) f=divtab[(z>>9)+DIVTAB_OFFSET_4- 256];
			else { dest++; /*printf("340K a\n");*/ goto skip_a; }
//			*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( ((v*f)>>20)&(64*63) ) | ( ((u*f)>>26)&(63) ) ] ];
//			*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( (uint) ( ((v*f)&aaa ) | ( ((uint)(u*f)>>6)) ) )>>20 ] ];
			*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( ((v*f)>>20)&(64*63) ) | ( (((uint)(u*f))>>26) ) ] ];

		  skip_a:
			l += dldx;
			u += dudx;
			v += dvdx;
			z += dzdx;
		}
#    else
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			int f;
			if (z<16384)  f=divtab[z>>1];
			else if (z<32768)  f=divtab[(z>>3)+DIVTAB_OFFSET_1-2048];
			else if (z<65536)  f=divtab[(z>>5)+DIVTAB_OFFSET_2-1024];
			else if (z<131072) f=divtab[(z>>7)+DIVTAB_OFFSET_3- 512];
			else if (z<320000) f=divtab[(z>>9)+DIVTAB_OFFSET_4- 256];
			else { dest++; goto skip_a; }
			*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( ( ((v>>6)*f)>>20)&(64*63) ) | ( ((u*f)>>26)&(63) ) ] ];
		  skip_a:
			l += dldx;
			u += dudx;
			v += dvdx;
			z += dzdx;
		}
#    endif
#  endif
#  ifdef DIV_TABLE_EXP /* Uses TABLE_1, slow, would have been elegant */
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			int f;
			int shifttab[8]={ 0,0,0,1,2,3,3,4 };

			if (z<32768) f=divtab[z>>6] -( (z&63)>>shifttab[z>>12]);
			else if (z<131072) f=divtab[z>>6];
			else if (z<320000) f=divtab[(z>>9) + DIVTAB_SIZE_0 -256];
//			else f=(65536*256*4+(z>>1))/(z) /*131*/; //65536*256*4/512000
			else { dest++; goto skip_a; }
			*dest++ = gr_fade_table[ (l&(0xff00)) + (uint)pixptr[ ( ( ((v>>6)*f)>>20)&(64*63) ) | ( ((u*f)>>26)&(63) ) ] ];
		  skip_a:
			l += dldx;
			u += dudx;
			v += dvdx;
			z += dzdx;
		}
#  endif
#endif /* USE_OLD_TMAP_SCANLINE: Y/N */
#endif /* OLD/Experimental version */

	} else {





/* PART 2 --------------------------------------------- */
#ifdef USE_OLD_TMAP_SCANLINE
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			c = (uint)pixptr[ ( (v/z)&(64*63) ) + ((u/z)&63) ];
			if ( c!=TRANSPARENCY_COLOR)
				*dest = gr_fade_table[ (l&(0xff00)) + c ];
			dest++;
			l += dldx;
			u += dudx;
			v += dvdx;
			z += dzdx;
		}
#else
#  ifdef DIV_TABLE_4
		for (x= fx_xright-fx_xleft+1 ; x > 0; --x ) {
			int f;
			if (z<16384)  f=divtab[z>>1];
			else if (z<32768)  f=divtab[(z>>3)+DIVTAB_OFFSET_1-2048];
			else if (z<65536)  f=divtab[(z>>5)+DIVTAB_OFFSET_2-1024];
			else if (z<131072) f=divtab[(z>>7)+DIVTAB_OFFSET_3- 512];
			else if (z<340000) f=divtab[(z>>9)+DIVTAB_OFFSET_4- 256];
			else { /*dest++;*/ /*printf("340K a\n");*/ goto skip_b; }
			c = (uint)pixptr[ ( ( (((v+32)>>6)*f)>>20)&(64*63) ) | ( ((u*f)>>26)&(63) ) ];
			if ( c!=TRANSPARENCY_COLOR)
				*dest = gr_fade_table[ (l&(0xff00)) + c ];
		  skip_b:
			dest++;
			l += dldx;
			u += dudx;
			v += dvdx;
			z += dzdx;
		}
#  endif /* DIV_TABLE_4 */
#endif /* USE_OLD_TMAP_SCANLINE: Y/N */
	}
}
#endif


void init_divtab() {
	int i;

#ifdef DIV_TABLE_4
	for (i=0; i<16384; i+=DIVTAB_STEP_0)
		divtab [i>>1] = (65536*256*4)/(i+DIVTAB_STEP_0/2);
	for (i=16384; i<32768; i+=DIVTAB_STEP_1)
		divtab [(i>>3)+DIVTAB_OFFSET_1- 16384/DIVTAB_STEP_1] = (65536*256*4)/(i+DIVTAB_STEP_1/2);
	for (i=32768; i<65536; i+=DIVTAB_STEP_2)
		divtab [(i>>5)+DIVTAB_OFFSET_2- 32768/DIVTAB_STEP_2] = (65536*256*4)/(i+DIVTAB_STEP_2/2);
	for (i=65536; i<131072; i+=DIVTAB_STEP_3)
		divtab [(i>>7)+DIVTAB_OFFSET_3- 65536/DIVTAB_STEP_3] = (65536*256*4)/(i+DIVTAB_STEP_2/2);
	for (i=131072; i<512*1024; i+=DIVTAB_STEP_4)
		divtab [(i>>9)+DIVTAB_OFFSET_4-131072/DIVTAB_STEP_4] = (65536*256*4)/(i+DIVTAB_STEP_4/2);
#endif
}
